from typing import Any, ClassVar

Maximization: OptimizationType
Minimization: OptimizationType

class IntegerConstraint:
    def __init__(self, *args, **kwargs) -> None: ...
    def check(self, *args, **kwargs) -> Any: ...
    @property
    def lb(self) -> Any: ...
    @property
    def ub(self) -> Any: ...

class IntegerSolution:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def x(self) -> Any: ...
    @property
    def y(self) -> Any: ...

class IntegerState:
    def __init__(self, arg0: IntegerSolution) -> None: ...
    @property
    def current(self) -> IntegerSolution: ...
    @property
    def current_best(self) -> IntegerSolution: ...
    @property
    def current_best_internal(self) -> IntegerSolution: ...
    @property
    def current_internal(self) -> IntegerSolution: ...
    @property
    def evaluations(self) -> int: ...
    @property
    def optimum_found(self) -> bool: ...

class LogInfo:
    def __init__(self, arg0: int, arg1: float, arg2: float, arg3: float, arg4: RealSolution, arg5: RealSolution) -> None: ...
    @property
    def current(self) -> RealSolution: ...
    @property
    def evaluations(self) -> int: ...
    @property
    def objective(self) -> RealSolution: ...
    @property
    def transformed_y(self) -> float: ...
    @property
    def transformed_y_best(self) -> float: ...
    @property
    def y_best(self) -> float: ...

class MetaData:
    def __init__(self, arg0: int, arg1: int, arg2: str, arg3: int, arg4: OptimizationType) -> None: ...
    @property
    def instance(self) -> int: ...
    @property
    def n_variables(self) -> int: ...
    @property
    def name(self) -> str: ...
    @property
    def optimization_type(self) -> Any: ...
    @property
    def problem_id(self) -> int: ...

class OptimizationType:
    __doc__: ClassVar[str] = ...  # read-only
    __members__: ClassVar[dict] = ...  # read-only
    Maximization: ClassVar[OptimizationType] = ...
    Minimization: ClassVar[OptimizationType] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class RealConstraint:
    def __init__(self, *args, **kwargs) -> None: ...
    def check(self, *args, **kwargs) -> Any: ...
    @property
    def lb(self) -> Any: ...
    @property
    def ub(self) -> Any: ...

class RealSolution:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def x(self) -> Any: ...
    @property
    def y(self) -> Any: ...

class RealState:
    def __init__(self, arg0: RealSolution) -> None: ...
    @property
    def current(self) -> RealSolution: ...
    @property
    def current_best(self) -> RealSolution: ...
    @property
    def current_best_internal(self) -> RealSolution: ...
    @property
    def current_internal(self) -> RealSolution: ...
    @property
    def evaluations(self) -> int: ...
    @property
    def optimum_found(self) -> bool: ...



import os
import random
import unittest
import shutil

import ioh

class Algorithm:
    def __init__(self):
        self.x = 10
        self.i = 1
    
    def __call__(self, p: ioh.problem.Real):
        for i in range(10000):
            x = list(map(lambda x: random.random(), range(p.meta_data.n_variables)))    
            p(x)
            self.i = i


class TestExperiment(unittest.TestCase):
    def test_experimenter(self):
        exp = ioh.Experiment(
            Algorithm(),
            [1], [1, 2], [5],
            njobs = 1,
            reps = 2,
            experiment_attributes = {"a": "1"},
            run_attributes = ['x'],
            logged_attributes = ['i']
        )

        def a_problem(x):
            return 0.0
            
        exp.add_custom_problem(a_problem, "Name")
        exp()

        info_files = {'IOHprofiler_f25_Name.info', 'IOHprofiler_f1_Sphere.info'}
        data_files = {'IOHprofiler_f25_DIM5.dat', 'IOHprofiler_f1_DIM5.dat'}

        for item in os.listdir('ioh_data'):
            path = os.path.join('ioh_data', item)
            if os.path.isfile(path) and item in info_files:
                self.assertNotEqual(os.path.getsize(path), 0)
                info_files.remove(item)
            elif os.path.isdir(path):
                for f in os.listdir(path):
                    if f in data_files:
                        path = os.path.join(path, f)
                        self.assertNotEqual(os.path.getsize(path), 0)
                        with open(path) as h:
                            data = list(filter(lambda x:x.startswith('"'), h))       

                        self.assertEqual(len(data), 4)
                        data_files.remove(f)

        self.assertSetEqual(info_files, set())
        self.assertSetEqual(data_files, set())

        self.assertTrue(os.path.isfile("ioh_data.zip"))

        shutil.rmtree("ioh_data")
        os.remove("ioh_data.zip")
           
    


if __name__ == "__main__":
    unittest.main()



from typing import Iterator, List

import ioh.iohcpp.logger

class BBOB(RealBase):
    def __init__(self, arg0: List[int], arg1: List[int], arg2: List[int]) -> None: ...

class Integer(IntegerBase):
    def __init__(self, arg0: List[int], arg1: List[int], arg2: List[int]) -> None: ...

class IntegerBase:
    def __init__(self, *args, **kwargs) -> None: ...
    def attach_logger(self, arg0: ioh.iohcpp.logger.Logger) -> None: ...
    def detach_logger(self) -> None: ...
    def reset(self) -> None: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    @property
    def dimensions(self) -> List[int]: ...
    @property
    def instances(self) -> List[int]: ...
    @property
    def name(self) -> str: ...
    @property
    def problem_ids(self) -> List[int]: ...

class PBO(IntegerBase):
    def __init__(self, arg0: List[int], arg1: List[int], arg2: List[int]) -> None: ...

class Real(RealBase):
    def __init__(self, arg0: List[int], arg1: List[int], arg2: List[int]) -> None: ...

class RealBase:
    def __init__(self, *args, **kwargs) -> None: ...
    def attach_logger(self, arg0: ioh.iohcpp.logger.Logger) -> None: ...
    def detach_logger(self) -> None: ...
    def reset(self) -> None: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    @property
    def dimensions(self) -> List[int]: ...
    @property
    def instances(self) -> List[int]: ...
    @property
    def name(self) -> str: ...
    @property
    def problem_ids(self) -> List[int]: ...



from typing import Tuple

import ioh.iohcpp
ALWAYS: Always
ON_IMPROVEMENT: OnImprovement

class Always(Trigger):
    def __init__(self) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class At(Trigger):
    def __init__(self, time_points: Set[int]) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    @property
    def time_points(self) -> Set[int]: ...

class During(Trigger):
    def __init__(self, time_ranges: Set[Tuple[int,int]]) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    @property
    def time_ranges(self) -> Set[Tuple[int,int]]: ...

class Each(Trigger):
    def __init__(self, interval: int, starting_at: int = ...) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    @property
    def interval(self) -> int: ...
    @property
    def starting_at(self) -> int: ...

class OnImprovement(Trigger):
    def __init__(self) -> None: ...
    def __getstate__(self) -> tuple: ...
    def __setstate__(self, arg0: tuple) -> None: ...

class Trigger:
    def __init__(self, *args, **kwargs) -> None: ...
    def reset(self) -> None: ...
    def __call__(self, arg0: ioh.iohcpp.LogInfo, arg1: ioh.iohcpp.MetaData) -> bool: ...



"""Python specific functions for IOH package

TODO: best far precision
TODO: Rename Integer -> Discrete
"""

import os
import math
import warnings
import itertools
import multiprocessing
import typing
import shutil
import copy

try:
    from .iohcpp import (
        problem,
        suite,
        logger,
        OptimizationType,
        RealSolution,
        IntegerSolution,
        IntegerConstraint,
        RealConstraint,
        RealState,
        IntegerState,
        MetaData,
        LogInfo,
    )
except ModuleNotFoundError:
    raise ModuleNotFoundError("No module named ioh")


ProblemType = typing.Union[problem.Real, problem.Integer]
VariableType = typing.Union[int, float]
ObjectiveType = typing.List[VariableType]


def get_problem(
    fid: typing.Union[int, str],
    instance: int = 1,
    dimension: int = 5,
    problem_type: str = "Real",
) -> ProblemType:
    """Instantiate a problem based on its function ID, dimension, instance and suite

    Parameters
    ----------
    fid: int or str
        The function ID of the problem in the suite, or the name of the function as string
    dimension: int
        The dimension (number of variables) of the problem
    instance: int
        The instance ID of the problem
    problem_type: str
        Which suite the problem is from. Either 'BBOB' or 'PBO' or 'Real' or 'Integer'
        Only used if fid is an int.

    """
    if (
        problem_type
        in (
            "PBO",
            "Integer",
        )
        and fid in [21, 23, "IsingTriangular", "NQueens"]
    ):
        if not math.sqrt(dimension).is_integer():
            raise ValueError(
                "For this function, the dimension needs to be a perfect square!"
            )

    bbob_fns = (
        getattr(problem, "BBOB").problems.values()
        | getattr(problem, "BBOB").problems.keys()
    )
    if (
        problem_type
        in (
            "BBOB",
            "Real",
        )
        and fid in bbob_fns
    ):
        if not dimension >= 2:
            raise ValueError("For BBOB functions the minimal dimension is 2")

    base_problem = getattr(problem, problem_type)
    if base_problem:
        if fid not in (base_problem.problems.values() | base_problem.problems.keys()):
            raise ValueError(
                f"{fid} is not registered for problem type: {problem_type}"
            )
        return base_problem.create(fid, instance, dimension)

    raise ValueError(f"Problem type {problem_type} is not supported")


def wrap_problem(
    function: typing.Callable[[ObjectiveType], float],
    name: str,
    problem_type: str,
    dimension: int = 5,
    instance: int = 1,
    optimization_type: OptimizationType = OptimizationType.Minimization,
    lb: VariableType = None,
    ub: VariableType = None,
    transform_variables: typing.Callable[[ObjectiveType, int], ObjectiveType] = None,
    transform_objectives: typing.Callable[[float, int], float] = None,
    calculate_objective: typing.Callable[
        [int, int], typing.Union[IntegerSolution, RealSolution]
    ] = None,
) -> ProblemType:
    """Function to wrap a callable as an ioh function

    Parameters
    ----------
    function: fn(x: list) -> float
        The callable to wrap
    name: str
        The name of the function. This can be used to create new instances of this function.
        Note, when not using unique names, this will override the previously wrapped functions.
    problem_type: str
        The type of the problem, accepted types are: 'Real' and 'Integer'
    dimension: int
        The dimension (number of variables) of the problem
    instance: int
        The instance ID of the problem
    optimization_type: OptimizationType
        The type of optimization to do, maximization or minimization
    lb: [int, float]
        The lower bound of the constraint, should be the same type as problem_type. When left to None, this will set
        to the lowest possible value for that type.
    ub: [int, float]
        The upper bound of the constraint, should be the same type as problem_type. When left to None, this will set
        to the highest possible value for that type.
    transform_variables: fn(x: list) -> list
        A function to transform the elements of x, prior to calling the function.
    transform_objectives: fn(y: float) -> float
        A function to tranform the float value of y, after callling the function.
    calculate_objective: fn(instance, dimension) -> Solution | (x, y)
        A function to calculate the global optimum of the function. This function gets a dimension and instance id,
        and should return either a Solution objective(IntegerSolution or RealSolution) or a tuple giving the
        x and y values for the global optimum. Where x is the search space representation and y the target value.
    """

    if problem_type == "Integer":
        wrapper = problem.wrap_integer_problem
    elif problem_type == "Real":
        wrapper = problem.wrap_real_problem
    else:
        raise ValueError(
            f"Problem type {problem_type} is not supported. Please select one of ('Integer', 'Real')"
        )

    wrapper(
        function,
        name,
        optimization_type,
        lb,
        ub,
        transform_variables,
        transform_objectives,
        calculate_objective,
    )
    return get_problem(name, instance, dimension, problem_type)


def get_problem_id(problem_name: str, problem_type: str):
    if problem_type in ("BBOB", "PBO", "Integer", "Real"):
        return {v: k for k, v in getattr(problem, problem_type).problems.items()}.get(
            problem_name
        )
    raise ValueError(f"Problem type {problem_type} is not supported")


class Experiment:
    def __init__(
        self,
        algorithm: typing.Any,
        fids: typing.List[int],
        iids: typing.List[int],
        dims: typing.List[int],
        reps: int = 1,
        problem_type: str = "Real",
        njobs: int = 1,
        logged: bool = True,
        logger_triggers: typing.List[logger.trigger.Trigger] = [
            logger.trigger.ON_IMPROVEMENT
        ],
        logger_additional_properties: typing.List[
            logger.property.AbstractProperty
        ] = [],
        output_directory: str = "./",
        folder_name: str = "ioh_data",
        algorithm_name: str = None,
        algorithm_info: str = "algorithm_info",
        store_positions: bool = False,
        experiment_attributes: typing.Dict[str, str] = {},
        run_attributes: typing.List[str] = [],
        logged_attributes: typing.List[str] = [],
        merge_output: bool = True,
        zip_output: bool = True,
        remove_data: bool = False,
    ):
        """
        Parameters
        ----------
            algorithm: typing.Any
                A instance of a given algorithm. Must implement a __call__ method
                taking a ioh.problem as argument.
            fids: typing.List[int]
                A list of integer problem ids to evaluate
            iids: typing.List[int],
                A list of integer problem instances to evaluate
            dims: typing.List[int],
                A list of integer problem dimensions to evaluate
            reps: int = 1,
                The number of independent repetitions for each problem, instance
                dimension combination
            problem_type: str = "BBOB"
                The type of problems to test. Available are "PBO", "BBOB",  "Integer"
                and "Real". Note that "Integer" and "Real" are supersets of "PBO"
                and "BBOB" resp.
            njobs: int = 1
                The number of parallel jobs, -1 assigns all available cpu's,
            logged: bool = True
                Whether or not the experiment uses a logger
            logger_triggers: typing.List[logger.trigger.Trigger]
                A list of trigger instances, used to determine when to trigger
            logger_additional_properties: typing.List[logger.property.AbstractProperty]
                A list of additional properties to be logged. If the properties to be logged are
                members of `algorithm`, you can also provide a list of attribute names to
                `logged_attributes`.
            output_directory: str = "./"
                The root output directory for the logger
            folder_name: str = "ioh_data"
                The name of the output directory for the logger
            algorithm_name: str = None
                A name for the algorithm. This is used in the log files.
            algorithm_info: str = ""
                Optional information, additional information used in log files
            optimization_type: OptimizationType = OptimizationType.Minimization
                The type of optimization
            store_positions: bool = False
                Whether to store the x-positions in the data-files
            experiment_attributes: typing.Dict[str, str] = []
                Attributes additionally stored per experiment. These are static
                throughout the experiment and stored in the .info files.
            run_attributes: typing.List[str] = []
                Names of attributes which are updated at every run, i.e. run index.
                These are stored in the .info files.
                Note that each of these attributes is required to be an attribute
                on algorithm.
            logged_attributes: typing.List[str] = []
                Names of attributes which are updated during the run of the algoritm,
                and are logged in the data files at every time point there is a
                logging action.
                Note that each of these attributes is required to be an attribute
                on algorithm.
            merge_output: bool = True
                Whether to merge output from multiple folders with the same
                folder_name. When run in parallel, seperate folder are produced.
                This option attempts to merge the folders back in single results.
                Note that this will also merge data already present in the folder,
                when the same name is used.
            zip_output: bool = True
                Whether to produce a .zip folder of output
            remove_data: bool = False
                Whether to remove all the produced data, except for the .zip file
                (when produced).
        """

        self.algorithm = algorithm
        self.logger_root = os.path.realpath(os.path.join(output_directory, folder_name))

        self.logger_params = dict(
            triggers=logger_triggers,
            additional_properties=logger_additional_properties,
            root=output_directory,
            folder_name=folder_name,
            algorithm_name=algorithm_name or str(algorithm),
            algorithm_info=algorithm_info,
            store_positions=store_positions,
        )
        self.fids = fids
        self.iids = iids
        self.dims = dims
        self.reps = reps
        self.logged = logged
        self.problem_type = problem_type
        self.njobs = njobs if njobs != -1 else multiprocessing.cpu_count()
        self.experiment_attributes = experiment_attributes
        self.logged_attributes = logged_attributes
        self.run_attributes = run_attributes
        self.merge_output = merge_output
        self.zip_output = zip_output
        self.remove_data = remove_data

        if os.path.isdir(self.logger_root) and self.merge_output:
            warnings.warn(
                f"There is already a folder named {self.logger_root} "
                f"and merge_output is set to True. The results of this experiment "
                f"will be merged with the contents of this folder",
                UserWarning,
            )

        for attr in itertools.chain(self.run_attributes, self.logged_attributes):
            if not hasattr(self.algorithm, attr):
                raise TypeError(
                    f"Attribute {attr} is not a member of algorithm {self.algorithm}"
                )

    def evaluate(self, ii: int, job: typing.Tuple[int, int, int]):
        """Evaluate a single function using self.algoritm.

        Note that this functions makes a copy of the algorithm for each new problem
        instantiation.

        Parameters
        ----------
        ii: int
            Iterator id
        job[fid: int, iid: int, dim: int]
            The problem id, the instance id and the problem dimension
        """

        algorithm = copy.deepcopy(self.algorithm)
        p = get_problem(*job, self.problem_type)
        if self.logged:
            logger_params = copy.deepcopy(self.logger_params)
            logger_params["folder_name"] += f"-tmp-{ii}"
            l = logger.Analyzer(**logger_params)
            l.set_experiment_attributes(self.experiment_attributes)
            l.add_run_attributes(algorithm, self.run_attributes)
            l.watch(algorithm, self.logged_attributes)
            l.reset()
            p.attach_logger(l)

        self.apply(algorithm, p)

    def apply(self, algorithm: any, problem: ProblemType):
        """Apply a given algorithm to a problem"""

        for _ in range(self.reps):
            algorithm(problem)
            problem.reset()

    def add_custom_problem(self, p: ProblemType, name: str = None, **kwargs):
        """Add a custom problem to the list of functions to be evaluated.

        Parameters
        ---------
        p: callable
            A function which takes an list of numeric values and return as list
            of numeric values

        name: str
            An optional name of the the newly added function.
        """

        name = name or "CustomProblem"

        if self.problem_type == "Integer":
            problem.wrap_integer_problem(p, name, **kwargs)
        elif self.problem_type == "Real":
            problem.wrap_real_problem(p, name, **kwargs)
        else:
            warnings.warn(
                "Adding custom problems is only allowed for problem_type in ('Integer', 'Real'). "
                "The added problem is ignored."
            )
            return

        self.fids.append(get_problem_id(name, self.problem_type))

    def merge_output_to_single_folder(self, prefix: str, target_folder: str):
        """Merges all ioh data folders into a single folder, having the same folder name prefix

        Parameters
        ----------
        prexix: str
            The prefix on which to select folders
        target_folder: str
            The target folder, i.e. the folder with the final output
        """

        root = os.path.dirname(self.logger_root)

        for folder_name in os.listdir(root):
            folder = os.path.join(root, folder_name)
            if not os.path.isdir(folder):
                continue

            if folder_name.startswith(prefix) and folder_name != os.path.basename(
                target_folder
            ):
                folder = os.path.join(root, folder_name)

                for info_file in os.listdir(folder):
                    source = os.path.join(folder, info_file)
                    if not os.path.isfile(source):
                        continue

                    if not info_file.endswith(".info"):
                        raise RuntimeError("Merging output with incompatible folders")

                    target = os.path.join(target_folder, info_file)
                    target_exists = os.path.isfile(target)
                    source_dat_folder = os.path.splitext(source)[0].replace(
                        "IOHprofiler", "data"
                    )
                    target_dat_folder = os.path.splitext(target)[0].replace(
                        "IOHprofiler", "data"
                    )

                    with open(source) as info_in, open(target, "a+") as info_out:
                        if target_exists:
                            info_out.write("\n")
                        for line in info_in:
                            info_out.write(line)
                    os.remove(source)

                    os.makedirs(target_dat_folder, exist_ok=True)
                    for dat_file in os.listdir(source_dat_folder):
                        source_dat_file = os.path.join(source_dat_folder, dat_file)
                        target_dat_file = os.path.join(target_dat_folder, dat_file)

                        with open(source_dat_file) as dat_in, open(
                            target_dat_file, "a+"
                        ) as dat_out:
                            for line in dat_in:
                                dat_out.write(line)

                        os.remove(source_dat_file)
                    os.removedirs(source_dat_folder)

    def merge_tmp_folders(self):
        """Merges the output of the temporary folders of the experiment"""

        root = os.path.basename(self.logger_root)
        target, idx = self.logger_root, 1
        while os.path.exists(target):
            target = root + f"-{idx}"
            idx += 1

        os.rename(self.logger_root + "-tmp-1", target)
        self.merge_output_to_single_folder(root + "-tmp", target)

    def run(self):
        """Alias for __call__"""

        return self()

    def __call__(self):
        """Run the experiment"""

        iterator = enumerate(itertools.product(self.fids, self.iids, self.dims), 1)
        if self.njobs != 1:
            with multiprocessing.Pool(self.njobs) as p:
                p.starmap(self.evaluate, iterator)
        else:
            for job in iterator:
                self.evaluate(*job)

        self.merge_tmp_folders()

        if self.merge_output:
            self.merge_output_to_single_folder(
                os.path.basename(self.logger_root), self.logger_root
            )

        if self.zip_output:
            shutil.make_archive(self.logger_root, "zip", self.logger_root)

        if self.remove_data:
            shutil.rmtree(self.logger_root)

        return self



from typing import Any, ClassVar, List, Optional

import ioh.iohcpp

class AbstractWModel(Integer):
    def __init__(self, problem_id: int, instance: int, n_variables: int, name: str, dummy_select_rate: float = ..., epistasis_block_size: int = ..., neutrality_mu: int = ..., ruggedness_gamma: int = ...) -> None: ...
    def wmodel_evaluate(self, arg0: List[int]) -> int: ...

class AttractiveSector(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class BBOB(Real):
    def __init__(self, *args, **kwargs) -> None: ...

class BentCigar(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class BuecheRastrigin(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class ConcatenatedTrap(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class DifferentPowers(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Discus(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Ellipsoid(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class EllipsoidRotated(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Gallagher101(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Gallagher21(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class GriewankRosenBrock(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Integer:
    problems: ClassVar[Any] = ...  # read-only
    def __init__(self, *args, **kwargs) -> None: ...
    def attach_logger(self, *args, **kwargs) -> Any: ...
    def create(self, *args, **kwargs) -> Any: ...
    def detach_logger(self, *args, **kwargs) -> Any: ...
    def reset(self, *args, **kwargs) -> Any: ...
    def __call__(self, *args, **kwargs) -> Any: ...
    @property
    def constraint(self) -> Any: ...
    @property
    def log_info(self) -> Any: ...
    @property
    def meta_data(self) -> Any: ...
    @property
    def objective(self) -> Any: ...
    @property
    def state(self) -> Any: ...

class IntegerWrappedProblem(Integer):
    def __init__(self, *args, **kwargs) -> None: ...

class IsingRing(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class IsingTorus(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class IsingTriangular(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Katsuura(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LABS(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnes(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesDummy1(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesDummy2(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesEpistasis(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesNeutrality(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesRuggedness1(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesRuggedness2(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LeadingOnesRuggedness3(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Linear(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LinearSlope(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class LunacekBiRastrigin(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class MIS(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class NQueens(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMax(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxDummy1(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxDummy2(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxEpistasis(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxNeutrality(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxRuggedness1(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxRuggedness2(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class OneMaxRuggedness3(Integer):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class PBO(Integer):
    def __init__(self, *args, **kwargs) -> None: ...

class Rastrigin(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class RastriginRotated(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Real:
    problems: ClassVar[Any] = ...  # read-only
    def __init__(self, *args, **kwargs) -> None: ...
    def attach_logger(self, *args, **kwargs) -> Any: ...
    def create(self, *args, **kwargs) -> Any: ...
    def detach_logger(self, *args, **kwargs) -> Any: ...
    def reset(self, *args, **kwargs) -> Any: ...
    def __call__(self, *args, **kwargs) -> Any: ...
    @property
    def constraint(self) -> Any: ...
    @property
    def log_info(self) -> Any: ...
    @property
    def meta_data(self) -> Any: ...
    @property
    def objective(self) -> Any: ...
    @property
    def state(self) -> Any: ...

class RealWrappedProblem(Real):
    def __init__(self, *args, **kwargs) -> None: ...

class Rosenbrock(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class RosenbrockRotated(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Schaffers10(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Schaffers1000(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Schwefel(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class SharpRidge(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class Sphere(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class StepEllipsoid(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

class WModelLeadingOnes(AbstractWModel):
    def __init__(self, instance: int, n_variables: int, dummy_select_rate: float = ..., epistasis_block_size: int = ..., neutrality_mu: int = ..., ruggedness_gamma: int = ...) -> None: ...

class WModelOneMax(AbstractWModel):
    def __init__(self, instance: int, n_variables: int, dummy_select_rate: float = ..., epistasis_block_size: int = ..., neutrality_mu: int = ..., ruggedness_gamma: int = ...) -> None: ...

class Weierstrass(Real):
    def __init__(self, arg0: int, arg1: int) -> None: ...

def wrap_integer_problem(f: handle, name: str, optimization_type: ioh.iohcpp.OptimizationType = ..., lb: Optional[float] = ..., ub: Optional[float] = ..., transform_variables: Optional[handle] = ..., transform_objectives: Optional[handle] = ..., calculate_objective: Optional[handle] = ...) -> None: ...
def wrap_real_problem(f: handle, name: str, optimization_type: ioh.iohcpp.OptimizationType = ..., lb: Optional[float] = ..., ub: Optional[float] = ..., transform_variables: Optional[handle] = ..., transform_objectives: Optional[handle] = ..., calculate_objective: Optional[handle] = ...) -> None: ...



import typing
from .iohcpp import IntegerConstraint as IntegerConstraint, IntegerSolution as IntegerSolution, IntegerState as IntegerState, LogInfo as LogInfo, MetaData as MetaData, OptimizationType, RealConstraint as RealConstraint, RealSolution as RealSolution, RealState as RealState, logger, suite as suite
from typing import Any

ProblemType: Any
VariableType: Any
ObjectiveType: Any

def get_problem(fid: typing.Union[int, str], instance: int = ..., dimension: int = ..., problem_type: str = ...) -> ProblemType: ...
def wrap_problem(function: typing.Callable[[ObjectiveType], float], name: str, problem_type: str, dimension: int = ..., instance: int = ..., optimization_type: OptimizationType = ..., lb: VariableType = ..., ub: VariableType = ..., transform_variables: typing.Callable[[ObjectiveType, int], ObjectiveType] = ..., transform_objectives: typing.Callable[[float, int], float] = ..., calculate_objective: typing.Callable[[int, int], typing.Union[IntegerSolution, RealSolution]] = ...) -> ProblemType: ...
def get_problem_id(problem_name: str, problem_type: str): ...

class Experiment:
    algorithm: Any
    logger_root: Any
    logger_params: Any
    fids: Any
    iids: Any
    dims: Any
    reps: Any
    logged: Any
    problem_type: Any
    njobs: Any
    experiment_attributes: Any
    logged_attributes: Any
    run_attributes: Any
    merge_output: Any
    zip_output: Any
    remove_data: Any
    def __init__(self, algorithm: typing.Any, fids: typing.List[int], iids: typing.List[int], dims: typing.List[int], reps: int = ..., problem_type: str = ..., njobs: int = ..., logged: bool = ..., logger_triggers: typing.List[logger.trigger.Trigger] = ..., logger_additional_properties: typing.List[logger.property.AbstractProperty] = ..., output_directory: str = ..., folder_name: str = ..., algorithm_name: str = ..., algorithm_info: str = ..., store_positions: bool = ..., experiment_attributes: typing.Dict[str, str] = ..., run_attributes: typing.List[str] = ..., logged_attributes: typing.List[str] = ..., merge_output: bool = ..., zip_output: bool = ..., remove_data: bool = ...) -> None: ...
    def evaluate(self, ii: int, job: typing.Tuple[int, int, int]): ...
    def apply(self, algorithm: any, problem: ProblemType): ...
    def add_custom_problem(self, p: ProblemType, name: str = ..., **kwargs): ...
    def merge_output_to_single_folder(self, prefix: str, target_folder: str): ...
    def merge_tmp_folders(self) -> None: ...
    def run(self): ...
    def __call__(self): ...



import unittest
import ioh
import os
import shutil
import pickle

class Container:
    def __init__(self):
        self.xy = 0
        self.xv = 10


class TestLogger(unittest.TestCase):
    def test_flatfile(self):
        pr = ioh.get_problem("Sphere", 1, 5) 
        p = [ioh.logger.property.CURRENT_Y_BEST]
        t = [ioh.logger.trigger.ON_IMPROVEMENT]
        l = ioh.logger.FlatFile(t, p, separator=" ")
        pr.attach_logger(l)
        self.assertEqual(os.path.realpath("./"), os.path.realpath(l.output_directory))
        pr([0] * 5)
        self.assertTrue(os.path.isfile("IOH.dat"))

    def test_analyzer(self):
        pr = ioh.get_problem("Sphere", 1, 5) 
        c = Container()
        l = ioh.logger.Analyzer([ioh.logger.trigger.ALWAYS])    
        l.set_experiment_attributes({"x":"1"})
        l.watch(c, ["xv", "xy"])
        l.add_run_attributes(c, "xv")
        pr.attach_logger(l)

        for c.xv in range(4):
            for i in range(5):
                c.xy = i*10
                pr([i] * 5)
            pr.reset()

        self.assertTrue(os.path.isdir("ioh_data"))

    def test_triggers(self):
        always = ioh.logger.trigger.Always()
        on_improvement = ioh.logger.trigger.OnImprovement()
        at = ioh.logger.trigger.At({1, 22})
        self.assertSetEqual(at.time_points, {1, 22})

        each = ioh.logger.trigger.Each(10, 2)
        self.assertEqual(each.interval, 10)
        self.assertEqual(each.starting_at, 2)

        during = ioh.logger.trigger.During({(10, 25), (30, 40)})
        self.assertEqual(during.time_ranges, {(10, 25), (30, 40)})

        log_info = ioh.LogInfo(22, 0, 0, 0, ioh.RealSolution([0, 1], 0), ioh.RealSolution([1, 0], 0))
        problem = ioh.get_problem(1, 1, 2)
        for t in (always, on_improvement, at, each, during):
            self.assertTrue(t(log_info, problem.meta_data))

    def tearDown(self) -> None:
        if os.path.isdir("ioh_data"):
            shutil.rmtree("ioh_data")
        if os.path.isfile("IOH.dat"):
            os.remove("IOH.dat")
        return super().tearDown()

    def test_pickle(self):
        try:
            pickle.dumps(ioh.logger.trigger.ON_IMPROVEMENT)
        except Exception as err: 
            self.fail(str(err))
        try:
            pickle.dumps(ioh.logger.property.RAW_Y_BEST)
        except Exception as err: 
            self.fail(str(err))
        try:
            pickle.dumps(ioh.logger.property.Property(Container(), "xy"))
        except Exception as err: 
            self.fail(str(err))

    def test_eaf(self):
        p = ioh.get_problem(1, 1, 5)
        l = ioh.logger.EAH(0, 20, 10, 0, 100, 10)
        self.assertEqual(l.error_range.min, 0)
        self.assertEqual(l.error_range.max, 20)
        self.assertEqual(l.error_range.size, 10)
        self.assertEqual(l.eval_range.min, 0)
        self.assertEqual(l.eval_range.max, 100)
        self.assertEqual(l.eval_range.size, 10)
        p.attach_logger(l)

        for i in range(10):
            p([i] * 5)

        self.assertEqual(list(l.data.keys()), [1])
                    
                    
if __name__ == "__main__":
    unittest.main()



import abc

class Problem(metaclass=abc.ABCMeta):

    # Objective function
    @abc.abstractmethod
    def call(self, *args ):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `call(self,*args):`')

    # i.e. for invariance checking
    @abc.abstractmethod
    def get_instance(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_instance(self)`')

    # Dimension
    @abc.abstractmethod
    def get_number_of_variables(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_number_of_variables(self)`')

    # Bounds
    @abc.abstractmethod
    def get_lowerbound(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_lowerbound(self)`')

    @abc.abstractmethod
    def get_upperbound(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `et_upperbound(self)`')


    # Generic information
    @abc.abstractmethod
    def get_minimization(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_minimization(self)`')

    @abc.abstractmethod
    def get_problem_id(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_problem_id(self)`')

    @abc.abstractmethod
    def get_name(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_name(self)`')

    @abc.abstractmethod
    def get_type(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_type(self)`')

    @abc.abstractmethod
    def get_number_of_objectives(self):
        raise NotImplementedError('Classes inheriting from ioh.Problem must define `get_number_of_objectives(self)`')




# -*- coding: utf-8 -*-
import os
import sys
import shutil
import platform
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}

if platform.system() == "Darwin":
    os.environ["CC"] = "clang"
    os.environ["CXX"] = "clang"
    os.environ["ARCHFLAGS"] = "-std=c++14"


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def generate_stubs(self):
        ext, *_ = self.extensions

        # remove any existing stubs
        for stubfile in os.listdir(os.path.join(ext.sourcedir, "ioh")):
            stubfile = os.path.join(ext.sourcedir, "ioh", stubfile)
            if stubfile.endswith(".pyi"):
                os.remove(stubfile)
        
        for subdir in ("iohcpp", "logger"):
            if os.path.isdir(os.path.join(ext.sourcedir, "ioh", subdir)):
                shutil.rmtree(os.path.join(ext.sourcedir, "ioh", subdir))

        # generate stub files
        command = """stubgen -m ioh -m ioh.iohcpp \
                -m ioh.iohcpp.problem \
                -m ioh.iohcpp.logger \
                -m ioh.iohcpp.suite \
                -m ioh.iohcpp.logger.trigger \
                -m ioh.iohcpp.logger.property \
                -o ./"""
        subprocess.check_call(command, cwd=ext.sourcedir, shell=True)

    def run(self):
        super().run()
        try:
            self.generate_stubs()
        except subprocess.CalledProcessError: 
            pass

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cfg = "Debug" if self.debug else "Release"
        print("build-type:", cfg)

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={}".format(extdir),
            "-DPYTHON_EXECUTABLE={}".format(sys.executable),
            "-DBUILD_PYTHON_PACKAGE=ON",
            "-DBUILD_TESTS=OFF",
            "-DBUILD_EXAMPLE=OFF",
            "-DEXAMPLE_VERSION_INFO={}".format(self.distribution.get_version()),
            "-DCMAKE_BUILD_TYPE={}".format(cfg),  # not used on MSVC, but no harm
        ]
        build_args = []

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator:
                cmake_args += ["-GNinja"]

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
                ]
                build_args += ["--config", cfg]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += ["-j{}".format(self.parallel)]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.

__version__ = "0.3.2.6"
gh_ref = os.environ.get("GITHUB_REF")
if gh_ref:
    *_, tag = gh_ref.split("/")
    __version__ = tag.replace("v", "")

with open("README.md", "r") as fh:
    long_description = fh.read()

iohcpp = CMakeExtension("ioh.iohcpp")

setup(
    name="ioh",
    version=__version__,
    author="Jacob de Nobel, Furong Ye, Diederick Vermetten, Hao Wang, Carola Doerr and Thomas Bäck",
    author_email="iohprofiler@liacs.leidenuniv.nl",
    description="The experimenter for Iterative Optimization Heuristics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    package_dir={'IOHexperimenter':'ioh'},
    package_data={"ioh": [
        "ioh/__init__.pyi",
        "ioh/iohcpp/__init__.pyi",
        "ioh/iohcpp/problem.pyi",
        "ioh/iohcpp/suite.pyi",
        "ioh/iohcpp/logger/__init__.pyi",
        "ioh/iohcpp/logger/property.pyi",
        "ioh/iohcpp/logger/trigger.pyi"
    ]},
    ext_modules=[iohcpp],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    test_suite='tests.python',
    python_requires='>=3.6',
    setup_requires=['cmake', 'ninja', 'pybind11', 'mypy']
)



import os
import unittest
import math

import ioh

DATA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "static")

class wmodel(ioh.problem.AbstractWModel):
    def __init__(self, instance, dim):
        super().__init__(100, instance, dim, "wmodel")

    def wmodel_evaluate(self, x) -> int:
        return x.count(1)


class TestProblem(unittest.TestCase):
    def test_get_problem(self):
        self.assertIsInstance(ioh.get_problem(1, 1, 2, "BBOB"), ioh.problem.Sphere)
        self.assertIsInstance(ioh.get_problem("Sphere", 1, 2, "BBOB"), ioh.problem.Sphere)
        self.assertIsInstance(ioh.get_problem("Sphere", 1, 2), ioh.problem.Sphere)
        self.assertIsInstance(ioh.get_problem(1, 1, 2, "PBO"), ioh.problem.OneMax)
        self.assertIsInstance(ioh.get_problem("OneMax", 1, 2, "PBO"), ioh.problem.OneMax)

    def test_wmodel(self):
        for p in map(lambda x: x(1, 10), (wmodel, ioh.problem.WModelLeadingOnes, ioh.problem.WModelOneMax)):
            self.assertEqual(p([1] * 10), 10)

           
    def test_evaluation_bbob_problems(self):
        for fid in range(1,25):
            f = ioh.get_problem(fid, 1 ,5, "BBOB")
            self.assertGreater(f([0,0,0,0,0]), -1000)

    def test_evaluation_pbo_problems(self):
        for fid in range(1,26):
            f = ioh.get_problem(fid, 1 ,4, "PBO")
            self.assertGreater(f([0,0,0,0]), -1000) 

    def test_bbob_problems_first_instance(self):
        expected = [
            161.17445568,
            12653420.412225708,
            -97.06158297486468,
            8845.53524810093,
            121.9133638725643,
            423021.00682286796,
            278.3290201933729,
            4315.032550201522,
            95168.253629631741,
            969025.0451803299,
            20911310.769634742,
            187251637.66430587,
            2198.7155122256763,
            261.60997479957,
            1103.1553337856833,
            273.87870697835791,
            35.410739487431208,
            107.64134358767826,
            -40.269198932753994,
            49739.05388887795,
            116.29585727504872,
            -914.902473409051,
            18.635078550302751,
            1782.2733296400438,
        ]
        for i in sorted(ioh.problem.BBOB.problems.keys())[:24]:
            p = ioh.problem.BBOB.create(i, 1, 5)
            self.assertTrue(
                math.isclose(p([0.1, 1., 2., 4., 5.4]), expected[i-1])
            )
   
    def test_pbo_problems_first_instance(self):
        expected = [
            5.0000,
            2.0000,
            24.0000,
            3.0000,
            4.0000,
            2.0000,
            6.0000,
            4.0000,
            6.0000,
            7.0000,
            3.0000,
            2.0000,
            1.0000,
            0.0000,
            2.0000,
            1.0000,
            1.0000,
            2.5312,
            5.0000,
            6.0000,
            9.0000,
            -4.0000,
            -16.0000,
            0.45,
            -0.70717,
        ]
        for i in sorted(ioh.problem.PBO.problems.keys())[:24]:
            p = ioh.problem.PBO.create(i, 1, 9)
            y = p([1, 1, 0, 1, 0, 0, 0, 1, 1])
            self.assertTrue(math.isclose(y, expected[i-1], abs_tol = 0.000099),
                msg=f"{p} expected: {expected[i-1]} got: {y}"
            )
   
    def test_file_comparisons(self):
        for test_file in  ("pbofitness16.in", "pbofitness100.in",
                           "bbobfitness5.in", "bbobfitness20.in", ):
            with self.subTest(test_file=test_file):
                suite, dim = test_file.split("fitness")
                dim = int(dim[:-3])
                dtype = float if suite == 'bbob' else int
                tol = .01 if suite == 'bbob' else 0.000099

                with open(os.path.join(DATA_DIR, test_file)) as f:
                    for line in f:
                        fid, iid, x, y = line.split()
                        if "," in x:
                            x = x.split(",")
                        x = list(map(dtype, x))
                        p = ioh.get_problem(int(fid), int(iid), dim, suite.upper())
                        self.assertTrue(math.isclose(p(x), float(y), abs_tol = tol))



if __name__ == "__main__":
    unittest.main()
